@@ -1,14 +1,30 @@
 import pandas as pd
-import numpy as np
 
-import datetime
+from functools import partial
+from itertools import product
 from contextlib import contextmanager
 
 from xbbg import __version__, const, pipeline
 from xbbg.io import logs, files, storage
 from xbbg.core import utils, conn, process
+from xbbg.core.conn import connect
 
-__all__ = ['__version__']
+__all__ = [
+    '__version__',
+    'connect',
+    'bdp',
+    'bds',
+    'bdh',
+    'bdib',
+    'bdtick',
+    'earning',
+    'dividend',
+    'beqs',
+    'live',
+    'subscribe',
+    'adjust_ccy',
+    'turnover',
+]
 
 
 def bdp(tickers, flds, **kwargs) -> pd.DataFrame:
@@ -28,9 +44,11 @@ def bdp(tickers, flds, **kwargs) -> pd.DataFrame:
     if isinstance(tickers, str): tickers = [tickers]
     if isinstance(flds, str): flds = [flds]
 
-    service = conn.bbg_service(service='//blp/refdata', **kwargs)
-    request = service.createRequest('ReferenceDataRequest')
-
+    request = process.create_request(
+        service='//blp/refdata',
+        request='ReferenceDataRequest',
+        **kwargs,
+    )
     process.init_request(request=request, tickers=tickers, flds=flds, **kwargs)
     logger.debug(f'Sending request to Bloomberg ...\n{request}')
     conn.send_request(request=request, **kwargs)
@@ -40,26 +58,25 @@ def bdp(tickers, flds, **kwargs) -> pd.DataFrame:
     if res.empty or any(fld not in res for fld in ['ticker', 'field']):
         return pd.DataFrame()
 
-    col_maps = kwargs.get('col_maps', None)
-    cols = res.field.unique()
     return (
         res
         .set_index(['ticker', 'field'])
         .unstack(level=1)
         .rename_axis(index=None, columns=[None, None])
         .droplevel(axis=1, level=0)
-        .loc[:, cols]
-        .pipe(pipeline.standard_cols, col_maps=col_maps)
+        .loc[:, res.field.unique()]
+        .pipe(pipeline.standard_cols, col_maps=kwargs.get('col_maps', None))
     )
 
 
-def bds(tickers, flds, **kwargs) -> pd.DataFrame:
+def bds(tickers, flds, use_port=False, **kwargs) -> pd.DataFrame:
     """
     Bloomberg block data
 
     Args:
         tickers: ticker(s)
         flds: field
+        use_port: use `PortfolioDataRequest`
         **kwargs: other overrides for query
 
     Returns:
@@ -67,43 +84,54 @@ def bds(tickers, flds, **kwargs) -> pd.DataFrame:
     """
     logger = logs.get_logger(bds, **kwargs)
 
-    service = conn.bbg_service(service='//blp/refdata', **kwargs)
-    request = service.createRequest('ReferenceDataRequest')
+    part = partial(_bds_, fld=flds, logger=logger, use_port=use_port, **kwargs)
+    if isinstance(tickers, str): tickers = [tickers]
+    return pd.DataFrame(pd.concat(map(part, tickers), sort=False))
 
-    if isinstance(tickers, str):
-        data_file = storage.ref_file(
-            ticker=tickers, fld=flds, has_date=True, ext='pkl', **kwargs
-        )
-        if files.exists(data_file):
-            logger.debug(f'Loading Bloomberg data from: {data_file}')
-            return pd.DataFrame(pd.read_pickle(data_file))
 
-        process.init_request(request=request, tickers=tickers, flds=flds, **kwargs)
-        logger.debug(f'Sending request to Bloomberg ...\n{request}')
-        conn.send_request(request=request, **kwargs)
+def _bds_(
+        ticker: str,
+        fld: str,
+        logger: logs.logging.Logger,
+        use_port: bool = False,
+        **kwargs,
+) -> pd.DataFrame:
+    """
+    Get data of BDS of single ticker
+    """
+    if 'has_date' not in kwargs: kwargs['has_date'] = True
+    data_file = storage.ref_file(ticker=ticker, fld=fld, ext='pkl', **kwargs)
+    if files.exists(data_file):
+        logger.debug(f'Loading Bloomberg data from: {data_file}')
+        return pd.DataFrame(pd.read_pickle(data_file))
 
-        res = pd.DataFrame(process.rec_events(func=process.process_ref, **kwargs))
-        if kwargs.get('raw', False): return res
-        if res.empty or any(fld not in res for fld in ['ticker', 'field']):
-            return pd.DataFrame()
+    request = process.create_request(
+        service='//blp/refdata',
+        request='PortfolioDataRequest' if use_port else 'ReferenceDataRequest',
+        **kwargs,
+    )
+    process.init_request(request=request, tickers=ticker, flds=fld, **kwargs)
+    logger.debug(f'Sending request to Bloomberg ...\n{request}')
+    conn.send_request(request=request, **kwargs)
 
-        data = (
-            res
-            .set_index(['ticker', 'field'])
-            .droplevel(axis=0, level=1)
-            .rename_axis(index=None)
-            .pipe(pipeline.standard_cols, col_maps=kwargs.get('col_maps', None))
-        )
-        if data_file:
-            logger.debug(f'Saving Bloomberg data to: {data_file}')
-            files.create_folder(data_file, is_file=True)
-            data.to_pickle(data_file)
-        return data
+    res = pd.DataFrame(process.rec_events(func=process.process_ref, **kwargs))
+    if kwargs.get('raw', False): return res
+    if res.empty or any(fld not in res for fld in ['ticker', 'field']):
+        return pd.DataFrame()
 
-    else:
-        return pd.DataFrame(pd.concat([
-            bds(tickers=ticker, flds=flds, **kwargs) for ticker in tickers
-        ], sort=False))
+    data = (
+        res
+        .set_index(['ticker', 'field'])
+        .droplevel(axis=0, level=1)
+        .rename_axis(index=None)
+        .pipe(pipeline.standard_cols, col_maps=kwargs.get('col_maps', None))
+    )
+    if data_file:
+        logger.debug(f'Saving Bloomberg data to: {data_file}')
+        files.create_folder(data_file, is_file=True)
+        data.to_pickle(data_file)
+
+    return data
 
 
 def bdh(
@@ -136,9 +164,11 @@ def bdh(
     if start_date is None: start_date = pd.Timestamp(e_dt) - pd.Timedelta(weeks=8)
     s_dt = utils.fmt_dt(start_date, fmt='%Y%m%d')
 
-    service = conn.bbg_service(service='//blp/refdata', **kwargs)
-    request = service.createRequest('HistoricalDataRequest')
-
+    request = process.create_request(
+        service='//blp/refdata',
+        request='HistoricalDataRequest',
+        **kwargs,
+    )
     process.init_request(
         request=request, tickers=tickers, flds=flds,
         start_date=s_dt, end_date=e_dt, adjust=adjust, **kwargs
@@ -162,9 +192,7 @@ def bdh(
     )
 
 
-def bdib(
-        ticker: str, dt, session='allday', typ='TRADE', **kwargs
-) -> pd.DataFrame:
+def bdib(ticker: str, dt, session='allday', typ='TRADE', **kwargs) -> pd.DataFrame:
     """
     Bloomberg intraday bar data
 
@@ -182,19 +210,16 @@ def bdib(
     Returns:
         pd.DataFrame
     """
-    from xbbg.core import missing
+    from xbbg.core import trials
 
     logger = logs.get_logger(bdib, **kwargs)
 
     ex_info = const.exch_info(ticker=ticker, **kwargs)
     if ex_info.empty: raise KeyError(f'Cannot find exchange info for {ticker}')
 
-    ss_rng = process.time_range(
-        dt=dt, ticker=ticker, session=session, tz=ex_info.tz, **kwargs
-    )
+    ss_rng = process.time_range(dt=dt, ticker=ticker, session=session, tz=ex_info.tz, **kwargs)
     data_file = storage.bar_file(ticker=ticker, dt=dt, typ=typ)
-    if files.exists(data_file) and kwargs.get('cache', True) \
-            and (not kwargs.get('reload', False)):
+    if files.exists(data_file) and kwargs.get('cache', True) and (not kwargs.get('reload', False)):
         res = (
             pd.read_parquet(data_file)
             .pipe(pipeline.add_ticker, ticker=ticker)
@@ -204,21 +229,11 @@ def bdib(
             logger.debug(f'Loading Bloomberg intraday data from: {data_file}')
             return res
 
-    t_1 = pd.Timestamp('today').date() - pd.Timedelta('1D')
-    whole_day = pd.Timestamp(dt).date() < t_1
-    batch = kwargs.pop('batch', False)
-    if (not whole_day) and batch:
-        logger.warning(f'Querying date {t_1} is too close, ignoring download ...')
-        return pd.DataFrame()
+    if not process.check_current(dt=dt, logger=logger, **kwargs): return pd.DataFrame()
 
     cur_dt = pd.Timestamp(dt).strftime('%Y-%m-%d')
-    info_log = f'{ticker} / {cur_dt} / {typ}'
-
     q_tckr = ticker
     if ex_info.get('is_fut', False):
-        if 'freq' not in ex_info:
-            logger.error(f'[freq] missing in info for {info_log} ...')
-
         is_sprd = ex_info.get('has_sprd', False) and (len(ticker[:-1]) != ex_info['tickers'][0])
         if not is_sprd:
             q_tckr = fut_ticker(gen_ticker=ticker, dt=dt, freq=ex_info['freq'])
@@ -227,32 +242,34 @@ def bdib(
                 return pd.DataFrame()
 
     info_log = f'{q_tckr} / {cur_dt} / {typ}'
-    miss_kw = dict(ticker=ticker, dt=dt, typ=typ, func='bdib')
-    cur_miss = missing.current_missing(**miss_kw)
-    if cur_miss >= 2:
-        if batch: return pd.DataFrame()
-        logger.info(f'{cur_miss} trials with no data {info_log}')
+    trial_kw = dict(ticker=ticker, dt=dt, typ=typ, func='bdib')
+    num_trials = trials.num_trials(**trial_kw)
+    if num_trials >= 2:
+        if kwargs.get('batch', False): return pd.DataFrame()
+        logger.info(f'{num_trials} trials with no data {info_log}')
         return pd.DataFrame()
 
-    service = conn.bbg_service(service='//blp/refdata', **kwargs)
-    request = service.createRequest('IntradayBarRequest')
-
     while conn.bbg_session(**kwargs).tryNextEvent(): pass
-    request.set('security', ticker)
-    request.set('eventType', typ)
-    request.set('interval', kwargs.get('interval', 1))
-
     time_rng = process.time_range(dt=dt, ticker=ticker, session='allday', **kwargs)
-    request.set('startDateTime', time_rng[0])
-    request.set('endDateTime', time_rng[1])
-
+    request = process.create_request(
+        service='//blp/refdata',
+        request='IntradayBarRequest',
+        settings=[
+            ('security', ticker),
+            ('eventType', typ),
+            ('interval', kwargs.get('interval', 1)),
+            ('startDateTime', time_rng[0]),
+            ('endDateTime', time_rng[1]),
+        ],
+        **kwargs,
+    )
     logger.debug(f'Sending request to Bloomberg ...\n{request}')
     conn.send_request(request=request, **kwargs)
 
     res = pd.DataFrame(process.rec_events(func=process.process_bar, **kwargs))
     if res.empty or ('time' not in res):
         logger.warning(f'No data for {info_log} ...')
-        missing.update_missing(**miss_kw)
+        trials.update_trials(cnt=num_trials + 1, **trial_kw)
         return pd.DataFrame()
 
     data = (
@@ -270,7 +287,7 @@ def bdib(
     return data.loc[ss_rng[0]:ss_rng[1]]
 
 
-def bdtick(ticker, dt, session='allday', types=None, **kwargs) -> pd.DataFrame:
+def bdtick(ticker, dt, session='allday', time_range=None, types=None, **kwargs) -> pd.DataFrame:
     """
     Bloomberg tick data
 
@@ -278,6 +295,8 @@ def bdtick(ticker, dt, session='allday', types=None, **kwargs) -> pd.DataFrame:
         ticker: ticker name
         dt: date to download
         session: [allday, day, am, pm, pre, post]
+        time_range: tuple of start and end time (must be converted into UTC)
+                    if this is given, `dt` and `session` will be ignored
         types: str or list, one or combinations of [
             TRADE, AT_TRADE, BID, ASK, MID_PRICE,
             BID_BEST, ASK_BEST, BEST_BID, BEST_ASK,
@@ -288,27 +307,44 @@ def bdtick(ticker, dt, session='allday', types=None, **kwargs) -> pd.DataFrame:
     """
     logger = logs.get_logger(bdtick, **kwargs)
 
+    if types is None: types = ['TRADE']
     exch = const.exch_info(ticker=ticker, **kwargs)
-    time_rng = process.time_range(dt=dt, ticker=ticker, session=session)
+    if exch.empty: raise LookupError(f'Cannot find exchange info for {ticker}')
 
-    service = conn.bbg_service(service='//blp/refdata', **kwargs)
-    request = service.createRequest('IntradayTickRequest')
+    if isinstance(time_range, (tuple, list)) and (len(time_range) == 2):
+        cur_dt = pd.Timestamp(dt).strftime('%Y-%m-%d')
+        time_rng = (
+            pd.DatetimeIndex([
+                f'{cur_dt} {time_range[0]}',
+                f'{cur_dt} {time_range[1]}',
+            ])
+            .tz_localize(exch.tz)
+            .tz_convert(process.DEFAULT_TZ)
+            .tz_convert('UTC')
+        )
+    else:
+        time_rng = process.time_range(dt=dt, ticker=ticker, session=session, **kwargs)
 
     while conn.bbg_session(**kwargs).tryNextEvent(): pass
-    if types is None: types = ['TRADE']
-    if isinstance(types, str): types = [types]
-    request.set('security', ticker)
-    for typ in types: request.append('eventTypes', typ)
-    request.set('startDateTime', time_rng[0])
-    request.set('endDateTime', time_rng[1])
-    request.set('includeConditionCodes', True)
-    request.set('includeExchangeCodes', True)
-    request.set('includeNonPlottableEvents', True)
-    request.set('includeBrokerCodes', True)
-    request.set('includeRpsCodes', True)
-    request.set('includeTradeTime', True)
-    request.set('includeActionCodes', True)
-    request.set('includeIndicatorCodes', True)
+    request = process.create_request(
+        service='//blp/refdata',
+        request='IntradayTickRequest',
+        settings=[
+            ('security', ticker),
+            ('startDateTime', time_rng[0]),
+            ('endDateTime', time_rng[1]),
+            ('includeConditionCodes', True),
+            ('includeExchangeCodes', True),
+            ('includeNonPlottableEvents', True),
+            ('includeBrokerCodes', True),
+            ('includeRpsCodes', True),
+            ('includeTradeTime', True),
+            ('includeActionCodes', True),
+            ('includeIndicatorCodes', True),
+        ],
+        append={'eventTypes': types},
+        **kwargs,
+    )
 
     logger.debug(f'Sending request to Bloomberg ...\n{request}')
     conn.send_request(request=request)
@@ -334,9 +370,7 @@ def bdtick(ticker, dt, session='allday', types=None, **kwargs) -> pd.DataFrame:
     )
 
 
-def earning(
-        ticker, by='Geo', typ='Revenue', ccy=None, level=None, **kwargs
-) -> pd.DataFrame:
+def earning(ticker, by='Geo', typ='Revenue', ccy=None, level=None, **kwargs) -> pd.DataFrame:
     """
     Earning exposures by Geo or Products
 
@@ -382,30 +416,12 @@ def earning(
 
     if 'level' not in data: raise KeyError('Cannot find [level] in data')
     for yr in data.columns[data.columns.str.startswith('fy')]:
-        pct = f'{yr}_pct'
-        data.loc[:, pct] = np.nan
-
-        # Calculate level 1 percentage
-        data.loc[data.level == 1, pct] = \
-            100 * data.loc[data.level == 1, yr] / data.loc[data.level == 1, yr].sum()
-
-        # Calculate level 2 percentage (higher levels will be ignored)
-        sub_pct = []
-        for r, snap in data.reset_index()[::-1].iterrows():
-            if snap.level > 2: continue
-            if snap.level == 1:
-                if len(sub_pct) == 0: continue
-                data.iloc[sub_pct, data.columns.get_loc(pct)] = \
-                    100 * data[yr].iloc[sub_pct] / data[yr].iloc[sub_pct].sum()
-                sub_pct = []
-            if snap.level == 2: sub_pct.append(r)
+        process.earning_pct(data=data, yr=yr)
 
     return data
 
 
-def dividend(
-        tickers, typ='all', start_date=None, end_date=None, **kwargs
-) -> pd.DataFrame:
+def dividend(tickers, typ='all', start_date=None, end_date=None, **kwargs) -> pd.DataFrame:
     """
     Bloomberg dividend / split history
 
@@ -433,52 +449,20 @@ def dividend(
     if isinstance(tickers, str): tickers = [tickers]
     tickers = [t for t in tickers if ('Equity' in t) and ('=' not in t)]
 
-    fld = {
-        'all': 'DVD_Hist_All',
-        'dvd': 'DVD_Hist',
-        'split': 'Eqy_DVD_Hist_Splits',
-        'gross': 'Eqy_DVD_Hist_Gross',
-        'adjust': 'Eqy_DVD_Adjust_Fact',
-        'adj_fund': 'Eqy_DVD_Adj_Fund',
-        'with_amt': 'DVD_Hist_All_with_Amt_Status',
-        'dvd_amt': 'DVD_Hist_with_Amt_Status',
-        'gross_amt': 'DVD_Hist_Gross_with_Amt_Stat',
-        'projected': 'BDVD_Pr_Ex_Dts_DVD_Amts_w_Ann',
-    }.get(typ, typ)
+    fld = const.DVD_TPYES.get(typ, typ)
 
     if (fld == 'Eqy_DVD_Adjust_Fact') and ('Corporate_Actions_Filter' not in kwargs):
         kwargs['Corporate_Actions_Filter'] = 'NORMAL_CASH|ABNORMAL_CASH|CAPITAL_CHANGE'
 
-    if fld in [
-        'DVD_Hist_All', 'DVD_Hist', 'Eqy_DVD_Hist_Gross',
-        'DVD_Hist_All_with_Amt_Status', 'DVD_Hist_with_Amt_Status',
-    ]:
-        if start_date:
-            kwargs['DVD_Start_Dt'] = utils.fmt_dt(start_date, fmt='%Y%m%d')
-        if end_date:
-            kwargs['DVD_End_Dt'] = utils.fmt_dt(end_date, fmt='%Y%m%d')
+    if start_date:
+        kwargs['DVD_Start_Dt'] = utils.fmt_dt(start_date, fmt='%Y%m%d')
+    if end_date:
+        kwargs['DVD_End_Dt'] = utils.fmt_dt(end_date, fmt='%Y%m%d')
 
-    return bds(tickers=tickers, flds=fld, col_maps={
-        'Declared Date': 'dec_date',
-        'Ex-Date': 'ex_date',
-        'Record Date': 'rec_date',
-        'Payable Date': 'pay_date',
-        'Dividend Amount': 'dvd_amt',
-        'Dividend Frequency': 'dvd_freq',
-        'Dividend Type': 'dvd_type',
-        'Amount Status': 'amt_status',
-        'Adjustment Date': 'adj_date',
-        'Adjustment Factor': 'adj_factor',
-        'Adjustment Factor Operator Type': 'adj_op',
-        'Adjustment Factor Flag': 'adj_flag',
-        'Amount Per Share': 'amt_ps',
-        'Projected/Confirmed': 'category',
-    }, **kwargs)
+    return bds(tickers=tickers, flds=fld, col_maps=const.DVD_COLS, **kwargs)
 
 
-def beqs(
-        screen, asof=None, typ='PRIVATE', group='General', **kwargs
-) -> pd.DataFrame:
+def beqs(screen, asof=None, typ='PRIVATE', group='General', **kwargs) -> pd.DataFrame:
     """
     Bloomberg equity screening
 
@@ -493,27 +477,24 @@ def beqs(
     """
     logger = logs.get_logger(beqs, **kwargs)
 
-    service = conn.bbg_service(service='//blp/refdata', **kwargs)
-    request = service.createRequest('BeqsRequest')
-
-    request.set('screenName', screen)
-    request.set('screenType', 'GLOBAL' if typ[0].upper() in ['G', 'B'] else 'PRIVATE')
-    request.set('Group', group)
-
-    if asof:
-        overrides = request.getElement('overrides')
-        ovrd = overrides.appendElement()
-        ovrd.setElement('fieldId', 'PiTDate')
-        ovrd.setElement('value', utils.fmt_dt(asof, '%Y%m%d'))
+    request = process.create_request(
+        service='//blp/refdata',
+        request='BeqsRequest',
+        settings=[
+            ('screenName', screen),
+            ('screenType', 'GLOBAL' if typ[0].upper() in ['G', 'B'] else 'PRIVATE'),
+            ('Group', group),
+        ],
+        ovrds=[('PiTDate', utils.fmt_dt(asof, '%Y%m%d'))] if asof else None,
+        **kwargs,
+    )
 
     logger.debug(f'Sending request to Bloomberg ...\n{request}')
     conn.send_request(request=request, **kwargs)
     res = pd.DataFrame(process.rec_events(func=process.process_ref, **kwargs))
     if res.empty:
         if kwargs.get('trial', 0): return pd.DataFrame()
-        else: return beqs(
-            screen=screen, asof=asof, typ=typ, group=group, trial=1, **kwargs
-        )
+        return beqs(screen=screen, asof=asof, typ=typ, group=group, trial=1, **kwargs)
 
     if kwargs.get('raw', False): return res
     cols = res.field.unique()
@@ -529,7 +510,7 @@ def beqs(
 
 
 @contextmanager
-def subscribe(tickers, flds=None, identity=None, **kwargs):
+def subscribe(tickers, flds=None, identity=None, options=None, **kwargs):
     """
     Subscribe Bloomberg realtime data
 
@@ -548,7 +529,7 @@ def subscribe(tickers, flds=None, identity=None, **kwargs):
         topic = f'//blp/mktdata/{ticker}'
         cid = conn.blpapi.CorrelationId(ticker)
         logger.debug(f'Subscribing {cid} => {topic}')
-        sub_list.add(topic, flds, correlationId=cid)
+        sub_list.add(topic, flds, correlationId=cid, options=options)
 
     try:
         conn.bbg_session(**kwargs).subscribe(sub_list, identity)
@@ -557,66 +538,63 @@ def subscribe(tickers, flds=None, identity=None, **kwargs):
         conn.bbg_session(**kwargs).unsubscribe(sub_list)
 
 
-def live(
-        tickers, flds='Last_Price', max_cnt=None, json=False, **kwargs
-) -> dict:
+async def live(tickers, flds=None, info=None, max_cnt=0, options=None, **kwargs):
     """
     Subscribe and getting data feeds from
 
     Args:
         tickers: list of tickers
         flds: fields to subscribe
+        info: list of keys of interests (ticker will be included)
         max_cnt: max number of data points to receive
-        json: if data is required to convert to json
 
     Yields:
         dict: Bloomberg market data
+
+    Examples:
+        >>> # async for _ in live('SPY US Equity', info=const.LIVE_INFO): pass
     """
+    from collections.abc import Iterable
+
     logger = logs.get_logger(live, **kwargs)
+    evt_typs = conn.event_types()
 
-    def get_value(element):
-        """
-        Get value from element
+    if flds is None:
+        s_flds = ['LAST_PRICE', 'BID', 'ASK']
+    else:
+        if isinstance(flds, str): flds = [flds]
+        s_flds = [fld.upper() for fld in flds]
 
-        Args:
-            element: Bloomberg element
+    if isinstance(info, str): info = [info]
+    if isinstance(info, Iterable): info = [key.upper() for key in info]
+    if info is None: info = const.LIVE_INFO
 
-        Returns:
-            dict
-        """
-        conv = [conn.blpapi.name.Name]
-        if json: conv += [pd.Timestamp, datetime.time, datetime.date]
-        if element.isNull(): return None
-        value = element.getValue()
-        if isinstance(value, np.bool_): return bool(value)
-        if isinstance(value, tuple(conv)): return str(value)
-        return value
-
-    if isinstance(flds, str): flds = [flds]
-    s_flds = [fld.upper() for fld in flds]
-    with subscribe(tickers=tickers, flds=s_flds, **kwargs):
+    sess = conn.bbg_session(**kwargs)
+    while sess.tryNextEvent(): pass
+    with subscribe(tickers=tickers, flds=s_flds, options=options, **kwargs):
         cnt = 0
-        while True if max_cnt is None else cnt < max_cnt:
+        while True and cnt <= max_cnt:
             try:
-                ev = conn.bbg_session(**kwargs).nextEvent(500)
-                if conn.event_types()[ev.eventType()] != 'SUBSCRIPTION_DATA':
-                    continue
-                for msg in ev:
-                    for fld in s_flds:
-                        if not msg.hasElement(fld): continue
-                        if msg.getElement(fld).isNull(): continue
-                        ticker = msg.correlationIds()[0].value()
-                        values = {**{'TICKER': ticker}, **{
-                            str(elem.name()): get_value(elem)
+                ev = sess.tryNextEvent()
+                if ev is None: continue
+                if evt_typs[ev.eventType()] != 'SUBSCRIPTION_DATA': continue
+
+                for msg, fld in product(ev, s_flds):
+                    if not msg.hasElement(fld): continue
+                    if msg.getElement(fld).isNull(): continue
+                    yield {
+                        **{
+                            'TICKER': msg.correlationIds()[0].value(),
+                            'FIELD': fld,
+                        },
+                        **{
+                            str(elem.name()): process.elem_value(elem)
                             for elem in msg.asElement().elements()
-                        }}
-                        yield {
-                            key: value for key, value in values.items()
-                            if value not in [np.nan, pd.NaT, None] or (
-                                isinstance(value, str) and value.strip()
-                            )
-                        }
-                        cnt += 1
+                            if (True if not info else str(elem.name()) in info)
+                        },
+                    }
+                    if max_cnt: cnt += 1
+
             except ValueError as e: logger.debug(e)
             except KeyboardInterrupt: break
 
@@ -624,9 +602,11 @@ def live(
 def active_futures(ticker: str, dt, **kwargs) -> str:
     """
     Active futures contract
+
     Args:
         ticker: futures ticker, i.e., ESA Index, Z A Index, CLA Comdty, etc.
         dt: date
+
     Returns:
         str: ticker name
     """
@@ -635,28 +615,28 @@ def active_futures(ticker: str, dt, **kwargs) -> str:
     info = const.market_info(f'{prefix[:-1]}1 {asset}')
 
     f1, f2 = f'{prefix[:-1]}1 {asset}', f'{prefix[:-1]}2 {asset}'
-    fut_2 = fut_ticker(gen_ticker=f2, dt=dt, freq=info['freq'], **kwargs)
-    fut_1 = fut_ticker(gen_ticker=f1, dt=dt, freq=info['freq'], **kwargs)
+    fut_2 = fut_ticker(gen_ticker=f2, dt=dt, freq=info.get('freq', 'M'), **kwargs)
+    fut_1 = fut_ticker(gen_ticker=f1, dt=dt, freq=info.get('freq', 'M'), **kwargs)
 
     fut_tk = bdp(tickers=[fut_1, fut_2], flds='Last_Tradeable_Dt')
 
     if pd.Timestamp(dt).month < pd.Timestamp(fut_tk.last_tradeable_dt[0]).month: return fut_1
 
     dts = pd.bdate_range(end=dt, periods=10)
-    volume = bdh(
-        fut_tk.index, flds='volume', start_date=dts[0], end_date=dts[-1], keep_one=True
-    )
+    volume = bdh(fut_tk.index, flds='volume', start_date=dts[0], end_date=dts[-1])
     if volume.empty: return fut_1
-    return volume.iloc[-1].idxmax()
+    return volume.iloc[-1].idxmax()[0]
 
 
 def fut_ticker(gen_ticker: str, dt, freq: str, **kwargs) -> str:
     """
     Get proper ticker from generic ticker
+
     Args:
         gen_ticker: generic ticker
         dt: date
         freq: futures contract frequency
+
     Returns:
         str: exact futures ticker
     """
@@ -712,3 +692,108 @@ def fut_ticker(gen_ticker: str, dt, freq: str, **kwargs) -> str:
     logger.debug(f'futures full chain:\n{fut_matu.to_string()}')
     logger.debug(f'getting index {idx} from:\n{sub_fut.to_string()}')
     return sub_fut.index.values[idx]
+
+
+def adjust_ccy(data: pd.DataFrame, ccy: str = 'USD') -> pd.DataFrame:
+    """
+    Adjust
+
+    Args:
+        data: daily price / turnover / etc. to adjust
+        ccy: currency to adjust to
+
+    Returns:
+        pd.DataFrame
+    """
+    if data.empty: return pd.DataFrame()
+    if ccy.lower() == 'local': return data
+    tickers = data.columns.get_level_values(level=0).unique()
+    start_date = data.index[0]
+    end_date = data.index[-1]
+
+    uccy = bdp(tickers=tickers, flds='crncy')
+    if not uccy.empty:
+        adj = (
+            uccy.crncy
+            .map(lambda v: {
+                'ccy': None if v.upper() == ccy else f'{ccy}{v.upper()} Curncy',
+                'factor': 100. if v[-1].islower() else 1.,
+            })
+            .apply(pd.Series)
+            .dropna(subset=['ccy'])
+        )
+    else: adj = pd.DataFrame()
+
+    if not adj.empty:
+        fx = (
+            bdh(tickers=adj.ccy.unique(), start_date=start_date, end_date=end_date)
+            .xs('Last_Price', axis=1, level=1)
+        )
+    else: fx = pd.DataFrame()
+
+    return (
+        pd.concat([
+            pd.Series(
+                (
+                    data[t]
+                    .dropna()
+                    .prod(axis=1)
+                    .div(
+                        (fx[adj.loc[t, 'ccy']] * adj.loc[t, 'factor'])
+                        if t in adj.index else 1.,
+                    )
+                ),
+                name=t,
+            )
+            for t in tickers
+        ], axis=1)
+    )
+
+
+def turnover(
+        tickers,
+        flds='Turnover',
+        start_date=None,
+        end_date=None,
+        ccy: str = 'USD',
+        factor: float = 1e6,
+) -> pd.DataFrame:
+    """
+    Currency adjusted turnover (in million)
+
+    Args:
+        tickers: ticker or list of tickers
+        flds: override `flds`,
+        start_date: start date, default 1 month prior to `end_date`
+        end_date: end date, default T - 1
+        ccy: currency - 'USD' (default), any currency, or 'local' (no adjustment)
+        factor: adjustment factor, default 1e6 - return values in millions
+
+    Returns:
+        pd.DataFrame
+    """
+    if end_date is None:
+        end_date = pd.bdate_range(end='today', periods=2)[0]
+    if start_date is None:
+        start_date = pd.bdate_range(end=end_date, periods=2, freq='M')[0]
+    if isinstance(tickers, str): tickers = [tickers]
+
+    data = bdh(tickers=tickers, flds=flds, start_date=start_date, end_date=end_date)
+    cols = data.columns.get_level_values(level=0).unique()
+
+    # If turnover is not available, use volume and vwap for calculation
+    use_volume = pd.DataFrame()
+    if isinstance(flds, str) and (flds.lower() == 'turnover'):
+        vol_tcks = [t for t in tickers if t not in cols]
+        if vol_tcks:
+            use_volume = turnover(
+                tickers=vol_tcks,
+                flds=['eqy_weighted_avg_px', 'volume'],
+                start_date=start_date,
+                end_date=end_date,
+                ccy=ccy,
+                factor=factor,
+            )
+
+    if data.empty and use_volume.empty: return pd.DataFrame()
+    return pd.concat([adjust_ccy(data=data, ccy=ccy).div(factor), use_volume], axis=1)
