@@ -20,6 +20,80 @@ ASSET_INFO = {
     'Equity': ['exch_codes'],
 }
 
+DVD_TPYES = {
+    'all': 'DVD_Hist_All',
+    'dvd': 'DVD_Hist',
+    'split': 'Eqy_DVD_Hist_Splits',
+    'gross': 'Eqy_DVD_Hist_Gross',
+    'adjust': 'Eqy_DVD_Adjust_Fact',
+    'adj_fund': 'Eqy_DVD_Adj_Fund',
+    'with_amt': 'DVD_Hist_All_with_Amt_Status',
+    'dvd_amt': 'DVD_Hist_with_Amt_Status',
+    'gross_amt': 'DVD_Hist_Gross_with_Amt_Stat',
+    'projected': 'BDVD_Pr_Ex_Dts_DVD_Amts_w_Ann',
+}
+
+DVD_COLS = {
+    'Declared Date': 'dec_date',
+    'Ex-Date': 'ex_date',
+    'Record Date': 'rec_date',
+    'Payable Date': 'pay_date',
+    'Dividend Amount': 'dvd_amt',
+    'Dividend Frequency': 'dvd_freq',
+    'Dividend Type': 'dvd_type',
+    'Amount Status': 'amt_status',
+    'Adjustment Date': 'adj_date',
+    'Adjustment Factor': 'adj_factor',
+    'Adjustment Factor Operator Type': 'adj_op',
+    'Adjustment Factor Flag': 'adj_flag',
+    'Amount Per Share': 'amt_ps',
+    'Projected/Confirmed': 'category',
+}
+
+LIVE_INFO = {
+    # Common fields
+    'MKTDATA_EVENT_TYPE', 'MKTDATA_EVENT_SUBTYPE', 'IS_DELAYED_STREAM',
+    # Last Price
+    'LAST_PRICE', 'RT_PX_CHG_PCT_1D', 'REALTIME_PERCENT_BID_ASK_SPREAD',
+    'EVT_TRADE_DATE_RT', 'TRADE_UPDATE_STAMP_RT',
+    'EQY_TURNOVER_REALTIME', 'VOLUME',
+    # Bid
+    'BID', 'BID_UPDATE_STAMP_RT',
+    # Ask
+    'ASK', 'ASK_UPDATE_STAMP_RT',
+    # Common in bid / ask
+    'SPREAD_BA', 'MID',
+}
+
+LIVE_CHG = {
+    'RT_PX_CHG_PCT_1D', 'CHG_PCT_1M_RT', 'CHG_PCT_3M_RT',
+    'CHG_PCT_MTD_RT', 'CHG_PCT_QTD_RT', 'CHG_PCT_YTD_RT',
+    'REALTIME_2_DAY_CHANGE_PERCENT', 'REALTIME_5_DAY_CHANGE_PERCENT',
+    'REALTIME_15_SEC_PRICE_PCT_CHG', 'REALTIME_ONE_MIN_PRICE_PCT_CHG',
+    # Equities only
+    'REALTIME_FIVE_MIN_PRICE_PCT_CHG', 'REALTIME_15_MIN_PRICE_PCT_CHG',
+    'REALTIME_ONE_HOUR_PRICE_PCT_CHG',
+}
+
+LIVE_VOL = {
+    'REALTIME_VOLUME_5_DAY_INTERVAL',
+    # Real-time current volume as % change from N-day avg volume
+    'DELTA_AVAT_1_DAY_INTERVAL', 'DELTA_AVAT_5_DAY_INTERVAL',
+    'DELTA_AVAT_10_DAY_INTERVAL', 'DELTA_AVAT_20_DAY_INTERVAL',
+    'DELTA_AVAT_30_DAY_INTERVAL', 'DELTA_AVAT_100_DAY_INTERVAL',
+    'DELTA_AVAT_180_DAY_INTERVAL',
+    # Real-time turnover as % change from N-day average turnover
+    'DELTA_ATAT_1_DAY_INTERVAL', 'DELTA_ATAT_5_DAY_INTERVAL',
+    'DELTA_ATAT_10_DAY_INTERVAL', 'DELTA_ATAT_20_DAY_INTERVAL',
+    'DELTA_ATAT_30_DAY_INTERVAL', 'DELTA_ATAT_100_DAY_INTERVAL',
+    'DELTA_ATAT_180_DAY_INTERVAL',
+}
+
+LIVE_RATIO = {
+    'PRICE_EARNINGS_RATIO_RT', 'PRICE_TO_BOOK_RATIO_RT',
+    'PRICE_TO_SALES_RATIO_RT', 'PRICE_CASH_FLOW_RT', 'PRICE_EBITDA_RT',
+}
+
 
 def exch_info(ticker: str, **kwargs) -> pd.Series:
     """
@@ -66,8 +140,8 @@ def exch_info(ticker: str, **kwargs) -> pd.Series:
         allday    [01:00, 21:00]
         day       [01:00, 21:00]
         Name: FuturesFinancialsICE, dtype: object
-        >>> exch_info('TESTTICKER Corp').empty
-        True
+        >>> exch_info('TESTTICKER Corp')
+        Series([], dtype: object)
         >>> exch_info('US')
         tz        America/New_York
         allday      [04:00, 20:00]
@@ -80,11 +154,15 @@ def exch_info(ticker: str, **kwargs) -> pd.Series:
         allday      [18:00, 17:00]
         day         [18:00, 17:00]
         Name: FuturesCBOE, dtype: object
+        >>> exch_info('TESTTICKER Index', original='TESTTICKER Index')
+        Series([], dtype: object)
+        >>> exch_info('TESTTCK Index')
+        Series([], dtype: object)
     """
     logger = logs.get_logger(exch_info, level='debug')
 
     if kwargs.get('ref', ''):
-        return exch_info(ticker=kwargs['ref'])
+        return exch_info(ticker=kwargs['ref'], **{k: v for k, v in kwargs.items() if k != 'ref'})
 
     exch = kwargs.get('config', param.load_config(cat='exch'))
     original = kwargs.get('original', '')
@@ -172,17 +250,17 @@ def market_info(ticker: str) -> pd.Series:
     #           Currency / Commodity / Index           #
     # ================================================ #
 
-    if t_info[-1] in ['Curncy', 'Comdty', 'Index']:
-        if t_info[0][-1].isdigit():
-            symbol = t_info[0][:-1].strip()
-            # Special contracts
-            if (symbol[:2] == 'UX') and (t_info[-1] == 'Index'):
-                symbol = 'UX'
-        else:
-            symbol = t_info[0].split('+')[0]
-        return take_first(data=a_info, query=f'tickers == "{symbol}"')
-
-    return pd.Series(dtype=object)
+    if t_info[0] in a_info.tickers.values:
+        symbol = t_info[0]
+    elif t_info[0][-1].isdigit():
+        end_idx = 2 if t_info[-2].isdigit() else 1
+        symbol = t_info[0][:-end_idx].strip()
+        # Special contracts
+        if (symbol[:2] == 'UX') and (t_info[-1] == 'Index'):
+            symbol = 'UX'
+    else:
+        symbol = t_info[0].split('+')[0]
+    return take_first(data=a_info, query=f'tickers == "{symbol}"')
 
 
 def take_first(data: pd.DataFrame, query: str) -> pd.Series:
@@ -221,7 +299,7 @@ def asset_config(asset: str) -> pd.DataFrame:
     config = (
         pd.concat([
             explode(
-                data=pd.DataFrame(param.load_yaml(cf)[asset]),
+                data=pd.DataFrame(param.load_yaml(cf).get(asset, [])),
                 columns=ASSET_INFO[asset],
             )
             for cf in cfg_files
@@ -245,6 +323,7 @@ def explode(data: pd.DataFrame, columns: list) -> pd.DataFrame:
     Returns:
         pd.DataFrame
     """
+    if data.empty: return pd.DataFrame()
     if len(columns) == 1:
         return data.explode(column=columns[0])
     return explode(
@@ -296,7 +375,7 @@ def ccy_pair(local, base='USD') -> CurrencyPair:
             info['factor'] *= 100.
 
     else:
-        logger = logs.get_logger(ccy_pair)
+        logger = logs.get_logger(ccy_pair, level='debug')
         logger.error(f'incorrect currency - local {local} / base {base}')
         return CurrencyPair(ticker='', factor=1., power=1.0)
 
@@ -334,7 +413,7 @@ def market_timing(ticker, dt, timing='EOD', tz='local', **kwargs) -> str:
         >>> market_timing('TESTTICKER Corp', dt='2018-09-10')
         ''
     """
-    logger = logs.get_logger(market_timing)
+    logger = logs.get_logger(market_timing, level='debug')
     exch = pd.Series(exch_info(ticker=ticker, **kwargs))
     if any(req not in exch.index for req in ['tz', 'allday', 'day']):
         logger.error(f'required exchange info cannot be found in {ticker} ...')
