@@ -1,8 +1,10 @@
 import pandas as pd
 import numpy as np
 
+from typing import Union
 
-def get_series(data: (pd.Series, pd.DataFrame), col='close') -> pd.DataFrame:
+
+def get_series(data: Union[pd.Series, pd.DataFrame], col='close') -> pd.DataFrame:
     """
     Get close column from intraday data
 
@@ -59,8 +61,9 @@ def standard_cols(data: pd.DataFrame, col_maps: dict = None) -> pd.DataFrame:
 
 
 def apply_fx(
-        data: (pd.Series, pd.DataFrame),
-        fx: (int, float, pd.Series, pd.DataFrame), power=-1.
+        data: Union[pd.Series, pd.DataFrame],
+        fx: Union[int, float, pd.Series, pd.DataFrame],
+        power=-1.,
 ) -> pd.DataFrame:
     """
     Apply FX to data
@@ -75,7 +78,7 @@ def apply_fx(
         where FX uses latest available price
 
     Examples:
-        >>> pd.set_option('precision', 2)
+        >>> pd.options.display.precision = 2
         >>> rms = (
         ...     pd.read_pickle('xbbg/tests/data/sample_rms_ib1.pkl')
         ...     .pipe(get_series, col='close')
@@ -116,12 +119,12 @@ def apply_fx(
     return data.mul(add_fx.iloc[:, -1].pow(power), axis=0).dropna(how='all')
 
 
-def daily_stats(data: (pd.Series, pd.DataFrame), **kwargs) -> pd.DataFrame:
+def daily_stats(data: Union[pd.Series, pd.DataFrame], **kwargs) -> pd.DataFrame:
     """
     Daily stats for given data
 
     Examples:
-        >>> pd.set_option('precision', 2)
+        >>> pd.options.display.precision = 2
         >>> (
         ...     pd.concat([
         ...         pd.read_pickle('xbbg/tests/data/sample_rms_ib0.pkl'),
@@ -140,8 +143,9 @@ def daily_stats(data: (pd.Series, pd.DataFrame), **kwargs) -> pd.DataFrame:
 
 
 def dropna(
-        data: (pd.Series, pd.DataFrame), cols: (int, list) = 0
-) -> (pd.Series, pd.DataFrame):
+        data: Union[pd.Series, pd.DataFrame],
+        cols: Union[int, list] = 0,
+) -> Union[pd.Series, pd.DataFrame]:
     """
     Drop NAs by columns
     """
@@ -228,9 +232,9 @@ def since_year(data: pd.DataFrame, year: int) -> pd.DataFrame:
         pd.DataFrame
 
     Examples:
-        >>> pd.set_option('display.width', 120)
-        >>> pd.set_option('display.max_columns', 10)
-        >>> pd.set_option('precision', 2)
+        >>> pd.options.display.width = 120
+        >>> pd.options.display.max_columns = 10
+        >>> pd.options.display.precision = 2
         >>> amzn = pd.read_pickle('xbbg/tests/data/sample_earning_amzn.pkl')
         >>> amzn.query('level == 1').pipe(since_year, year=2017)
                          segment_name  level    fy2018    fy2017  fy2018_pct  fy2017_pct
@@ -248,7 +252,7 @@ def since_year(data: pd.DataFrame, year: int) -> pd.DataFrame:
     )]
 
 
-def perf(data: pd.DataFrame) -> pd.DataFrame:
+def perf(data: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
     """
     Performance rebased to 100
 
@@ -267,17 +271,14 @@ def perf(data: pd.DataFrame) -> pd.DataFrame:
         3  103.0  104.0
         4   99.0  110.0
     """
-    return pd.DataFrame(
-        pd.concat([
-            (
-                srs
-                .dropna()
-                .pct_change()
-                .fillna(0)
-                .add(1)
-                .cumprod()
-                .mul(100)
-            )
-            for _, srs in data.items()
-        ], axis=1)
-    )
+    if isinstance(data, pd.Series):
+        return (
+            data
+            .dropna()
+            .pct_change()
+            .fillna(0)
+            .add(1)
+            .cumprod()
+            .mul(100)
+        )
+    return data.apply(perf, axis=0)
